if __name__ == '__main__':
	import forge
else:
	import ctypes

	class Context:
		def __init__(self, name = "Context"):
#			print "Context.init " + name
			self.objName = name
			self.verbose = False
			self.nameCount = {}

			terminalDL = ctypes.CDLL("libfexTerminalDL.so")
			self.fe_terminal_execute = terminalDL.fe_terminal_execute

		def setVerbose(self, verbose):
			self.verbose = verbose

		def execute(self, command):
			size = 256
			result = ctypes.create_string_buffer(size)

			if self.verbose:
				print "\n[36m" + command + "[0m"

			self.fe_terminal_execute(size,result,command)

			if self.verbose:
				print "[34m" + result.value + "[0m"

			return result.value

		def name(self):
			return self.objName

		def setFrame(self, frame):
			command = "set frame " + str(frame)
			self.execute(command)

		def manage(self, library):
#			print "Context.manage " + library

			command = "manage " + library
			self.execute(command)

		def load(self, path, name = False, options = ""):
			name = name or path

#			print "Context.load " + path + " name " + name

			command = "load %s %s \"%s\"" % (name, path, options)
			self.execute(command)

			surface = Surface(self, name)
			return surface

		def create(self, operator, name = False):
			name = name or operator

#			print "Context.create " + operator + " name " + name

			command = "create " + operator + " " + name
			self.execute(command)

			operator = Operator(self, name)
			return operator

	class Operator:
		def __init__(self, context, name = "Operator"):
#			print "Operator.init " + name
			self.context = context
			self.objName = name

		def __del__(self):
#			print "Operator.del " + self.objName
			self.release()

		def name(self):
			return self.objName

		def operate(self, input0 = None, input1 = None, input2 = None, input3 = None, name = False):
			name = name or (self.objName + ".out")

#			print "Operator.operate " + name

			command = "signal " + self.objName + " " + name

			if input0:
#				print "  input0 " + input0.name()
				command += " " + input0.name()
			if input1:
#				print "  input1 " + input1.name()
				command += " " + input1.name()
			if input2:
#				print "  input2 " + input2.name()
				command += " " + input2.name()
			if input3:
#				print "  input3 " + input3.name()
				command += " " + input3.name()

			self.context.execute(command)

			surface = Surface(self.context, name)
			return surface

		def set(self, attribute, value, typename = False):
			if not typename:
				if type(value) is float:
					typename = "real"
				elif type(value) is int:
					typename = "integer"
				elif type(value) is bool:
					typename = "boolean"
				elif type(value) is str:
					typename = "string"
				elif type(value) is list and len(value) == 3:
					typename = "vector3"
					value = "\"%G %G %G\"" % (value[0], value[1], value[2])
				else:
					print "Operator.set incompatible type " + str(type(value))
					return

#			print "Operator.set \"" + attribute + "\" " + typename + " " + str(value)

			command = "set %s \"%s\" %s %s" % (self.objName, attribute, typename, str(value))
			self.context.execute(command)

		def get(self, attribute):
#			print "Operator.get " + attribute

			command = "get %s \"%s\"" % (self.objName, attribute)
			return self.context.execute(command)

		def __call__(self, input0 = None, input1 = None, input2 = None, input3 = None, name = False):
			return self.operate(input0, input1, input2, input3)

		def __setitem__(self, attribute, value):
			self.set(attribute, value)

		def __getitem__(self, attribute):
			return self.get(attribute)

		def release(self):
#			print "Operator.release"

			command = "release " + self.objName
			self.context.execute(command)

	class Surface:
		def __init__(self, context, name = "Surface"):
#			print "Surface.init " + name
			self.context = context
			self.objName = name

			if not name in context.nameCount:
				context.nameCount[name] = 0

			context.nameCount[name]+=1

		def __del__(self):
#			print "Surface.del " + self.objName
			self.release()

		def name(self):
			return self.objName

		def save(self, path, options = ""):
#			print "Surface.save " + path

			command = "save %s %s \"%s\"" % (self.objName, path, options)
			self.context.execute(command)

		def release(self):
#			print "Surface.release"

			self.context.nameCount[self.objName]-=1

			if self.context.nameCount[self.objName] == 0:

				command = "release " + self.objName
				self.context.execute(command)
